from __future__ import annotations

from typing import TYPE_CHECKING

from puffling.metadata.custom import CustomMetadataHook
from puffling.plugin import hookimpl

if TYPE_CHECKING:
    from puffling.metadata.plugin.interface import MetadataHookInterface


@hookimpl
def puffling_register_metadata_hook() -> type[MetadataHookInterface]:
    return CustomMetadataHook  # type: ignore



import json
import pathlib
import time
from contextlib import closing
from io import StringIO

import httpx

LATEST_API = "https://api.github.com/repos/spdx/license-list-data/releases/latest"
LICENSES_URL = (
    "https://raw.githubusercontent.com/spdx/license-list-data/v{}/json/licenses.json"
)
EXCEPTIONS_URL = (
    "https://raw.githubusercontent.com/spdx/license-list-data/v{}/json/exceptions.json"
)


def download_data(url):
    for _ in range(600):
        try:
            response = httpx.get(url)
            response.raise_for_status()
        except Exception:  # noqa: BLE001
            time.sleep(1)
            continue
        else:
            return json.loads(response.content.decode("utf-8"))

    message = "Download failed"
    raise ConnectionError(message)


def main():
    latest_version = download_data(LATEST_API)["tag_name"][1:]

    licenses = {}
    for license_data in download_data(LICENSES_URL.format(latest_version))["licenses"]:
        license_id = license_data["licenseId"]
        deprecated = license_data["isDeprecatedLicenseId"]
        licenses[license_id.lower()] = {"id": license_id, "deprecated": deprecated}

    exceptions = {}
    for exception_data in download_data(EXCEPTIONS_URL.format(latest_version))[
        "exceptions"
    ]:
        exception_id = exception_data["licenseExceptionId"]
        deprecated = exception_data["isDeprecatedLicenseId"]
        exceptions[exception_id.lower()] = {
            "id": exception_id,
            "deprecated": deprecated,
        }

    project_root = pathlib.Path(__file__).resolve().parent.parent
    data_file = project_root / "src" / "puffling" / "licenses" / "supported.py"

    with closing(StringIO()) as file_contents:
        file_contents.write(
            f"""\
from __future__ import annotations

VERSION = {latest_version!r}\n\nLICENSES: dict[str, dict[str, str | bool]] = {{
"""
        )

        for normalized_name, data in sorted(licenses.items()):
            file_contents.write(f"    {normalized_name!r}: {data!r},\n")

        file_contents.write("}\n\nEXCEPTIONS: dict[str, dict[str, str | bool]] = {\n")

        for normalized_name, data in sorted(exceptions.items()):
            file_contents.write(f"    {normalized_name!r}: {data!r},\n")

        file_contents.write("}\n")

        with data_file.open("w", encoding="utf-8") as f:
            f.write(file_contents.getvalue())


if __name__ == "__main__":
    main()



from __future__ import annotations

VERSION = "3.19"

LICENSES: dict[str, dict[str, str | bool]] = {
    "0bsd": {"id": "0BSD", "deprecated": False},
    "aal": {"id": "AAL", "deprecated": False},
    "abstyles": {"id": "Abstyles", "deprecated": False},
    "adobe-2006": {"id": "Adobe-2006", "deprecated": False},
    "adobe-glyph": {"id": "Adobe-Glyph", "deprecated": False},
    "adsl": {"id": "ADSL", "deprecated": False},
    "afl-1.1": {"id": "AFL-1.1", "deprecated": False},
    "afl-1.2": {"id": "AFL-1.2", "deprecated": False},
    "afl-2.0": {"id": "AFL-2.0", "deprecated": False},
    "afl-2.1": {"id": "AFL-2.1", "deprecated": False},
    "afl-3.0": {"id": "AFL-3.0", "deprecated": False},
    "afmparse": {"id": "Afmparse", "deprecated": False},
    "agpl-1.0": {"id": "AGPL-1.0", "deprecated": True},
    "agpl-1.0-only": {"id": "AGPL-1.0-only", "deprecated": False},
    "agpl-1.0-or-later": {"id": "AGPL-1.0-or-later", "deprecated": False},
    "agpl-3.0": {"id": "AGPL-3.0", "deprecated": True},
    "agpl-3.0-only": {"id": "AGPL-3.0-only", "deprecated": False},
    "agpl-3.0-or-later": {"id": "AGPL-3.0-or-later", "deprecated": False},
    "aladdin": {"id": "Aladdin", "deprecated": False},
    "amdplpa": {"id": "AMDPLPA", "deprecated": False},
    "aml": {"id": "AML", "deprecated": False},
    "ampas": {"id": "AMPAS", "deprecated": False},
    "antlr-pd": {"id": "ANTLR-PD", "deprecated": False},
    "antlr-pd-fallback": {"id": "ANTLR-PD-fallback", "deprecated": False},
    "apache-1.0": {"id": "Apache-1.0", "deprecated": False},
    "apache-1.1": {"id": "Apache-1.1", "deprecated": False},
    "apache-2.0": {"id": "Apache-2.0", "deprecated": False},
    "apafml": {"id": "APAFML", "deprecated": False},
    "apl-1.0": {"id": "APL-1.0", "deprecated": False},
    "app-s2p": {"id": "App-s2p", "deprecated": False},
    "apsl-1.0": {"id": "APSL-1.0", "deprecated": False},
    "apsl-1.1": {"id": "APSL-1.1", "deprecated": False},
    "apsl-1.2": {"id": "APSL-1.2", "deprecated": False},
    "apsl-2.0": {"id": "APSL-2.0", "deprecated": False},
    "arphic-1999": {"id": "Arphic-1999", "deprecated": False},
    "artistic-1.0": {"id": "Artistic-1.0", "deprecated": False},
    "artistic-1.0-cl8": {"id": "Artistic-1.0-cl8", "deprecated": False},
    "artistic-1.0-perl": {"id": "Artistic-1.0-Perl", "deprecated": False},
    "artistic-2.0": {"id": "Artistic-2.0", "deprecated": False},
    "baekmuk": {"id": "Baekmuk", "deprecated": False},
    "bahyph": {"id": "Bahyph", "deprecated": False},
    "barr": {"id": "Barr", "deprecated": False},
    "beerware": {"id": "Beerware", "deprecated": False},
    "bitstream-vera": {"id": "Bitstream-Vera", "deprecated": False},
    "bittorrent-1.0": {"id": "BitTorrent-1.0", "deprecated": False},
    "bittorrent-1.1": {"id": "BitTorrent-1.1", "deprecated": False},
    "blessing": {"id": "blessing", "deprecated": False},
    "blueoak-1.0.0": {"id": "BlueOak-1.0.0", "deprecated": False},
    "borceux": {"id": "Borceux", "deprecated": False},
    "bsd-1-clause": {"id": "BSD-1-Clause", "deprecated": False},
    "bsd-2-clause": {"id": "BSD-2-Clause", "deprecated": False},
    "bsd-2-clause-freebsd": {"id": "BSD-2-Clause-FreeBSD", "deprecated": True},
    "bsd-2-clause-netbsd": {"id": "BSD-2-Clause-NetBSD", "deprecated": True},
    "bsd-2-clause-patent": {"id": "BSD-2-Clause-Patent", "deprecated": False},
    "bsd-2-clause-views": {"id": "BSD-2-Clause-Views", "deprecated": False},
    "bsd-3-clause": {"id": "BSD-3-Clause", "deprecated": False},
    "bsd-3-clause-attribution": {"id": "BSD-3-Clause-Attribution", "deprecated": False},
    "bsd-3-clause-clear": {"id": "BSD-3-Clause-Clear", "deprecated": False},
    "bsd-3-clause-lbnl": {"id": "BSD-3-Clause-LBNL", "deprecated": False},
    "bsd-3-clause-modification": {
        "id": "BSD-3-Clause-Modification",
        "deprecated": False,
    },
    "bsd-3-clause-no-military-license": {
        "id": "BSD-3-Clause-No-Military-License",
        "deprecated": False,
    },
    "bsd-3-clause-no-nuclear-license": {
        "id": "BSD-3-Clause-No-Nuclear-License",
        "deprecated": False,
    },
    "bsd-3-clause-no-nuclear-license-2014": {
        "id": "BSD-3-Clause-No-Nuclear-License-2014",
        "deprecated": False,
    },
    "bsd-3-clause-no-nuclear-warranty": {
        "id": "BSD-3-Clause-No-Nuclear-Warranty",
        "deprecated": False,
    },
    "bsd-3-clause-open-mpi": {"id": "BSD-3-Clause-Open-MPI", "deprecated": False},
    "bsd-4-clause": {"id": "BSD-4-Clause", "deprecated": False},
    "bsd-4-clause-shortened": {"id": "BSD-4-Clause-Shortened", "deprecated": False},
    "bsd-4-clause-uc": {"id": "BSD-4-Clause-UC", "deprecated": False},
    "bsd-protection": {"id": "BSD-Protection", "deprecated": False},
    "bsd-source-code": {"id": "BSD-Source-Code", "deprecated": False},
    "bsl-1.0": {"id": "BSL-1.0", "deprecated": False},
    "busl-1.1": {"id": "BUSL-1.1", "deprecated": False},
    "bzip2-1.0.5": {"id": "bzip2-1.0.5", "deprecated": True},
    "bzip2-1.0.6": {"id": "bzip2-1.0.6", "deprecated": False},
    "c-uda-1.0": {"id": "C-UDA-1.0", "deprecated": False},
    "cal-1.0": {"id": "CAL-1.0", "deprecated": False},
    "cal-1.0-combined-work-exception": {
        "id": "CAL-1.0-Combined-Work-Exception",
        "deprecated": False,
    },
    "caldera": {"id": "Caldera", "deprecated": False},
    "catosl-1.1": {"id": "CATOSL-1.1", "deprecated": False},
    "cc-by-1.0": {"id": "CC-BY-1.0", "deprecated": False},
    "cc-by-2.0": {"id": "CC-BY-2.0", "deprecated": False},
    "cc-by-2.5": {"id": "CC-BY-2.5", "deprecated": False},
    "cc-by-2.5-au": {"id": "CC-BY-2.5-AU", "deprecated": False},
    "cc-by-3.0": {"id": "CC-BY-3.0", "deprecated": False},
    "cc-by-3.0-at": {"id": "CC-BY-3.0-AT", "deprecated": False},
    "cc-by-3.0-de": {"id": "CC-BY-3.0-DE", "deprecated": False},
    "cc-by-3.0-igo": {"id": "CC-BY-3.0-IGO", "deprecated": False},
    "cc-by-3.0-nl": {"id": "CC-BY-3.0-NL", "deprecated": False},
    "cc-by-3.0-us": {"id": "CC-BY-3.0-US", "deprecated": False},
    "cc-by-4.0": {"id": "CC-BY-4.0", "deprecated": False},
    "cc-by-nc-1.0": {"id": "CC-BY-NC-1.0", "deprecated": False},
    "cc-by-nc-2.0": {"id": "CC-BY-NC-2.0", "deprecated": False},
    "cc-by-nc-2.5": {"id": "CC-BY-NC-2.5", "deprecated": False},
    "cc-by-nc-3.0": {"id": "CC-BY-NC-3.0", "deprecated": False},
    "cc-by-nc-3.0-de": {"id": "CC-BY-NC-3.0-DE", "deprecated": False},
    "cc-by-nc-4.0": {"id": "CC-BY-NC-4.0", "deprecated": False},
    "cc-by-nc-nd-1.0": {"id": "CC-BY-NC-ND-1.0", "deprecated": False},
    "cc-by-nc-nd-2.0": {"id": "CC-BY-NC-ND-2.0", "deprecated": False},
    "cc-by-nc-nd-2.5": {"id": "CC-BY-NC-ND-2.5", "deprecated": False},
    "cc-by-nc-nd-3.0": {"id": "CC-BY-NC-ND-3.0", "deprecated": False},
    "cc-by-nc-nd-3.0-de": {"id": "CC-BY-NC-ND-3.0-DE", "deprecated": False},
    "cc-by-nc-nd-3.0-igo": {"id": "CC-BY-NC-ND-3.0-IGO", "deprecated": False},
    "cc-by-nc-nd-4.0": {"id": "CC-BY-NC-ND-4.0", "deprecated": False},
    "cc-by-nc-sa-1.0": {"id": "CC-BY-NC-SA-1.0", "deprecated": False},
    "cc-by-nc-sa-2.0": {"id": "CC-BY-NC-SA-2.0", "deprecated": False},
    "cc-by-nc-sa-2.0-fr": {"id": "CC-BY-NC-SA-2.0-FR", "deprecated": False},
    "cc-by-nc-sa-2.0-uk": {"id": "CC-BY-NC-SA-2.0-UK", "deprecated": False},
    "cc-by-nc-sa-2.5": {"id": "CC-BY-NC-SA-2.5", "deprecated": False},
    "cc-by-nc-sa-3.0": {"id": "CC-BY-NC-SA-3.0", "deprecated": False},
    "cc-by-nc-sa-3.0-de": {"id": "CC-BY-NC-SA-3.0-DE", "deprecated": False},
    "cc-by-nc-sa-3.0-igo": {"id": "CC-BY-NC-SA-3.0-IGO", "deprecated": False},
    "cc-by-nc-sa-4.0": {"id": "CC-BY-NC-SA-4.0", "deprecated": False},
    "cc-by-nd-1.0": {"id": "CC-BY-ND-1.0", "deprecated": False},
    "cc-by-nd-2.0": {"id": "CC-BY-ND-2.0", "deprecated": False},
    "cc-by-nd-2.5": {"id": "CC-BY-ND-2.5", "deprecated": False},
    "cc-by-nd-3.0": {"id": "CC-BY-ND-3.0", "deprecated": False},
    "cc-by-nd-3.0-de": {"id": "CC-BY-ND-3.0-DE", "deprecated": False},
    "cc-by-nd-4.0": {"id": "CC-BY-ND-4.0", "deprecated": False},
    "cc-by-sa-1.0": {"id": "CC-BY-SA-1.0", "deprecated": False},
    "cc-by-sa-2.0": {"id": "CC-BY-SA-2.0", "deprecated": False},
    "cc-by-sa-2.0-uk": {"id": "CC-BY-SA-2.0-UK", "deprecated": False},
    "cc-by-sa-2.1-jp": {"id": "CC-BY-SA-2.1-JP", "deprecated": False},
    "cc-by-sa-2.5": {"id": "CC-BY-SA-2.5", "deprecated": False},
    "cc-by-sa-3.0": {"id": "CC-BY-SA-3.0", "deprecated": False},
    "cc-by-sa-3.0-at": {"id": "CC-BY-SA-3.0-AT", "deprecated": False},
    "cc-by-sa-3.0-de": {"id": "CC-BY-SA-3.0-DE", "deprecated": False},
    "cc-by-sa-4.0": {"id": "CC-BY-SA-4.0", "deprecated": False},
    "cc-pddc": {"id": "CC-PDDC", "deprecated": False},
    "cc0-1.0": {"id": "CC0-1.0", "deprecated": False},
    "cddl-1.0": {"id": "CDDL-1.0", "deprecated": False},
    "cddl-1.1": {"id": "CDDL-1.1", "deprecated": False},
    "cdl-1.0": {"id": "CDL-1.0", "deprecated": False},
    "cdla-permissive-1.0": {"id": "CDLA-Permissive-1.0", "deprecated": False},
    "cdla-permissive-2.0": {"id": "CDLA-Permissive-2.0", "deprecated": False},
    "cdla-sharing-1.0": {"id": "CDLA-Sharing-1.0", "deprecated": False},
    "cecill-1.0": {"id": "CECILL-1.0", "deprecated": False},
    "cecill-1.1": {"id": "CECILL-1.1", "deprecated": False},
    "cecill-2.0": {"id": "CECILL-2.0", "deprecated": False},
    "cecill-2.1": {"id": "CECILL-2.1", "deprecated": False},
    "cecill-b": {"id": "CECILL-B", "deprecated": False},
    "cecill-c": {"id": "CECILL-C", "deprecated": False},
    "cern-ohl-1.1": {"id": "CERN-OHL-1.1", "deprecated": False},
    "cern-ohl-1.2": {"id": "CERN-OHL-1.2", "deprecated": False},
    "cern-ohl-p-2.0": {"id": "CERN-OHL-P-2.0", "deprecated": False},
    "cern-ohl-s-2.0": {"id": "CERN-OHL-S-2.0", "deprecated": False},
    "cern-ohl-w-2.0": {"id": "CERN-OHL-W-2.0", "deprecated": False},
    "checkmk": {"id": "checkmk", "deprecated": False},
    "clartistic": {"id": "ClArtistic", "deprecated": False},
    "cnri-jython": {"id": "CNRI-Jython", "deprecated": False},
    "cnri-python": {"id": "CNRI-Python", "deprecated": False},
    "cnri-python-gpl-compatible": {
        "id": "CNRI-Python-GPL-Compatible",
        "deprecated": False,
    },
    "coil-1.0": {"id": "COIL-1.0", "deprecated": False},
    "community-spec-1.0": {"id": "Community-Spec-1.0", "deprecated": False},
    "condor-1.1": {"id": "Condor-1.1", "deprecated": False},
    "copyleft-next-0.3.0": {"id": "copyleft-next-0.3.0", "deprecated": False},
    "copyleft-next-0.3.1": {"id": "copyleft-next-0.3.1", "deprecated": False},
    "cpal-1.0": {"id": "CPAL-1.0", "deprecated": False},
    "cpl-1.0": {"id": "CPL-1.0", "deprecated": False},
    "cpol-1.02": {"id": "CPOL-1.02", "deprecated": False},
    "crossword": {"id": "Crossword", "deprecated": False},
    "crystalstacker": {"id": "CrystalStacker", "deprecated": False},
    "cua-opl-1.0": {"id": "CUA-OPL-1.0", "deprecated": False},
    "cube": {"id": "Cube", "deprecated": False},
    "curl": {"id": "curl", "deprecated": False},
    "d-fsl-1.0": {"id": "D-FSL-1.0", "deprecated": False},
    "diffmark": {"id": "diffmark", "deprecated": False},
    "dl-de-by-2.0": {"id": "DL-DE-BY-2.0", "deprecated": False},
    "doc": {"id": "DOC", "deprecated": False},
    "dotseqn": {"id": "Dotseqn", "deprecated": False},
    "drl-1.0": {"id": "DRL-1.0", "deprecated": False},
    "dsdp": {"id": "DSDP", "deprecated": False},
    "dvipdfm": {"id": "dvipdfm", "deprecated": False},
    "ecl-1.0": {"id": "ECL-1.0", "deprecated": False},
    "ecl-2.0": {"id": "ECL-2.0", "deprecated": False},
    "ecos-2.0": {"id": "eCos-2.0", "deprecated": True},
    "efl-1.0": {"id": "EFL-1.0", "deprecated": False},
    "efl-2.0": {"id": "EFL-2.0", "deprecated": False},
    "egenix": {"id": "eGenix", "deprecated": False},
    "elastic-2.0": {"id": "Elastic-2.0", "deprecated": False},
    "entessa": {"id": "Entessa", "deprecated": False},
    "epics": {"id": "EPICS", "deprecated": False},
    "epl-1.0": {"id": "EPL-1.0", "deprecated": False},
    "epl-2.0": {"id": "EPL-2.0", "deprecated": False},
    "erlpl-1.1": {"id": "ErlPL-1.1", "deprecated": False},
    "etalab-2.0": {"id": "etalab-2.0", "deprecated": False},
    "eudatagrid": {"id": "EUDatagrid", "deprecated": False},
    "eupl-1.0": {"id": "EUPL-1.0", "deprecated": False},
    "eupl-1.1": {"id": "EUPL-1.1", "deprecated": False},
    "eupl-1.2": {"id": "EUPL-1.2", "deprecated": False},
    "eurosym": {"id": "Eurosym", "deprecated": False},
    "fair": {"id": "Fair", "deprecated": False},
    "fdk-aac": {"id": "FDK-AAC", "deprecated": False},
    "frameworx-1.0": {"id": "Frameworx-1.0", "deprecated": False},
    "freebsd-doc": {"id": "FreeBSD-DOC", "deprecated": False},
    "freeimage": {"id": "FreeImage", "deprecated": False},
    "fsfap": {"id": "FSFAP", "deprecated": False},
    "fsful": {"id": "FSFUL", "deprecated": False},
    "fsfullr": {"id": "FSFULLR", "deprecated": False},
    "fsfullrwd": {"id": "FSFULLRWD", "deprecated": False},
    "ftl": {"id": "FTL", "deprecated": False},
    "gd": {"id": "GD", "deprecated": False},
    "gfdl-1.1": {"id": "GFDL-1.1", "deprecated": True},
    "gfdl-1.1-invariants-only": {"id": "GFDL-1.1-invariants-only", "deprecated": False},
    "gfdl-1.1-invariants-or-later": {
        "id": "GFDL-1.1-invariants-or-later",
        "deprecated": False,
    },
    "gfdl-1.1-no-invariants-only": {
        "id": "GFDL-1.1-no-invariants-only",
        "deprecated": False,
    },
    "gfdl-1.1-no-invariants-or-later": {
        "id": "GFDL-1.1-no-invariants-or-later",
        "deprecated": False,
    },
    "gfdl-1.1-only": {"id": "GFDL-1.1-only", "deprecated": False},
    "gfdl-1.1-or-later": {"id": "GFDL-1.1-or-later", "deprecated": False},
    "gfdl-1.2": {"id": "GFDL-1.2", "deprecated": True},
    "gfdl-1.2-invariants-only": {"id": "GFDL-1.2-invariants-only", "deprecated": False},
    "gfdl-1.2-invariants-or-later": {
        "id": "GFDL-1.2-invariants-or-later",
        "deprecated": False,
    },
    "gfdl-1.2-no-invariants-only": {
        "id": "GFDL-1.2-no-invariants-only",
        "deprecated": False,
    },
    "gfdl-1.2-no-invariants-or-later": {
        "id": "GFDL-1.2-no-invariants-or-later",
        "deprecated": False,
    },
    "gfdl-1.2-only": {"id": "GFDL-1.2-only", "deprecated": False},
    "gfdl-1.2-or-later": {"id": "GFDL-1.2-or-later", "deprecated": False},
    "gfdl-1.3": {"id": "GFDL-1.3", "deprecated": True},
    "gfdl-1.3-invariants-only": {"id": "GFDL-1.3-invariants-only", "deprecated": False},
    "gfdl-1.3-invariants-or-later": {
        "id": "GFDL-1.3-invariants-or-later",
        "deprecated": False,
    },
    "gfdl-1.3-no-invariants-only": {
        "id": "GFDL-1.3-no-invariants-only",
        "deprecated": False,
    },
    "gfdl-1.3-no-invariants-or-later": {
        "id": "GFDL-1.3-no-invariants-or-later",
        "deprecated": False,
    },
    "gfdl-1.3-only": {"id": "GFDL-1.3-only", "deprecated": False},
    "gfdl-1.3-or-later": {"id": "GFDL-1.3-or-later", "deprecated": False},
    "giftware": {"id": "Giftware", "deprecated": False},
    "gl2ps": {"id": "GL2PS", "deprecated": False},
    "glide": {"id": "Glide", "deprecated": False},
    "glulxe": {"id": "Glulxe", "deprecated": False},
    "glwtpl": {"id": "GLWTPL", "deprecated": False},
    "gnuplot": {"id": "gnuplot", "deprecated": False},
    "gpl-1.0": {"id": "GPL-1.0", "deprecated": True},
    "gpl-1.0+": {"id": "GPL-1.0+", "deprecated": True},
    "gpl-1.0-only": {"id": "GPL-1.0-only", "deprecated": False},
    "gpl-1.0-or-later": {"id": "GPL-1.0-or-later", "deprecated": False},
    "gpl-2.0": {"id": "GPL-2.0", "deprecated": True},
    "gpl-2.0+": {"id": "GPL-2.0+", "deprecated": True},
    "gpl-2.0-only": {"id": "GPL-2.0-only", "deprecated": False},
    "gpl-2.0-or-later": {"id": "GPL-2.0-or-later", "deprecated": False},
    "gpl-2.0-with-autoconf-exception": {
        "id": "GPL-2.0-with-autoconf-exception",
        "deprecated": True,
    },
    "gpl-2.0-with-bison-exception": {
        "id": "GPL-2.0-with-bison-exception",
        "deprecated": True,
    },
    "gpl-2.0-with-classpath-exception": {
        "id": "GPL-2.0-with-classpath-exception",
        "deprecated": True,
    },
    "gpl-2.0-with-font-exception": {
        "id": "GPL-2.0-with-font-exception",
        "deprecated": True,
    },
    "gpl-2.0-with-gcc-exception": {
        "id": "GPL-2.0-with-GCC-exception",
        "deprecated": True,
    },
    "gpl-3.0": {"id": "GPL-3.0", "deprecated": True},
    "gpl-3.0+": {"id": "GPL-3.0+", "deprecated": True},
    "gpl-3.0-only": {"id": "GPL-3.0-only", "deprecated": False},
    "gpl-3.0-or-later": {"id": "GPL-3.0-or-later", "deprecated": False},
    "gpl-3.0-with-autoconf-exception": {
        "id": "GPL-3.0-with-autoconf-exception",
        "deprecated": True,
    },
    "gpl-3.0-with-gcc-exception": {
        "id": "GPL-3.0-with-GCC-exception",
        "deprecated": True,
    },
    "gsoap-1.3b": {"id": "gSOAP-1.3b", "deprecated": False},
    "haskellreport": {"id": "HaskellReport", "deprecated": False},
    "hippocratic-2.1": {"id": "Hippocratic-2.1", "deprecated": False},
    "hpnd": {"id": "HPND", "deprecated": False},
    "hpnd-sell-variant": {"id": "HPND-sell-variant", "deprecated": False},
    "htmltidy": {"id": "HTMLTIDY", "deprecated": False},
    "ibm-pibs": {"id": "IBM-pibs", "deprecated": False},
    "icu": {"id": "ICU", "deprecated": False},
    "ijg": {"id": "IJG", "deprecated": False},
    "imagemagick": {"id": "ImageMagick", "deprecated": False},
    "imatix": {"id": "iMatix", "deprecated": False},
    "imlib2": {"id": "Imlib2", "deprecated": False},
    "info-zip": {"id": "Info-ZIP", "deprecated": False},
    "intel": {"id": "Intel", "deprecated": False},
    "intel-acpi": {"id": "Intel-ACPI", "deprecated": False},
    "interbase-1.0": {"id": "Interbase-1.0", "deprecated": False},
    "ipa": {"id": "IPA", "deprecated": False},
    "ipl-1.0": {"id": "IPL-1.0", "deprecated": False},
    "isc": {"id": "ISC", "deprecated": False},
    "jam": {"id": "Jam", "deprecated": False},
    "jasper-2.0": {"id": "JasPer-2.0", "deprecated": False},
    "jpnic": {"id": "JPNIC", "deprecated": False},
    "json": {"id": "JSON", "deprecated": False},
    "knuth-ctan": {"id": "Knuth-CTAN", "deprecated": False},
    "lal-1.2": {"id": "LAL-1.2", "deprecated": False},
    "lal-1.3": {"id": "LAL-1.3", "deprecated": False},
    "latex2e": {"id": "Latex2e", "deprecated": False},
    "leptonica": {"id": "Leptonica", "deprecated": False},
    "lgpl-2.0": {"id": "LGPL-2.0", "deprecated": True},
    "lgpl-2.0+": {"id": "LGPL-2.0+", "deprecated": True},
    "lgpl-2.0-only": {"id": "LGPL-2.0-only", "deprecated": False},
    "lgpl-2.0-or-later": {"id": "LGPL-2.0-or-later", "deprecated": False},
    "lgpl-2.1": {"id": "LGPL-2.1", "deprecated": True},
    "lgpl-2.1+": {"id": "LGPL-2.1+", "deprecated": True},
    "lgpl-2.1-only": {"id": "LGPL-2.1-only", "deprecated": False},
    "lgpl-2.1-or-later": {"id": "LGPL-2.1-or-later", "deprecated": False},
    "lgpl-3.0": {"id": "LGPL-3.0", "deprecated": True},
    "lgpl-3.0+": {"id": "LGPL-3.0+", "deprecated": True},
    "lgpl-3.0-only": {"id": "LGPL-3.0-only", "deprecated": False},
    "lgpl-3.0-or-later": {"id": "LGPL-3.0-or-later", "deprecated": False},
    "lgpllr": {"id": "LGPLLR", "deprecated": False},
    "libpng": {"id": "Libpng", "deprecated": False},
    "libpng-2.0": {"id": "libpng-2.0", "deprecated": False},
    "libselinux-1.0": {"id": "libselinux-1.0", "deprecated": False},
    "libtiff": {"id": "libtiff", "deprecated": False},
    "libutil-david-nugent": {"id": "libutil-David-Nugent", "deprecated": False},
    "liliq-p-1.1": {"id": "LiLiQ-P-1.1", "deprecated": False},
    "liliq-r-1.1": {"id": "LiLiQ-R-1.1", "deprecated": False},
    "liliq-rplus-1.1": {"id": "LiLiQ-Rplus-1.1", "deprecated": False},
    "linux-man-pages-copyleft": {"id": "Linux-man-pages-copyleft", "deprecated": False},
    "linux-openib": {"id": "Linux-OpenIB", "deprecated": False},
    "lpl-1.0": {"id": "LPL-1.0", "deprecated": False},
    "lpl-1.02": {"id": "LPL-1.02", "deprecated": False},
    "lppl-1.0": {"id": "LPPL-1.0", "deprecated": False},
    "lppl-1.1": {"id": "LPPL-1.1", "deprecated": False},
    "lppl-1.2": {"id": "LPPL-1.2", "deprecated": False},
    "lppl-1.3a": {"id": "LPPL-1.3a", "deprecated": False},
    "lppl-1.3c": {"id": "LPPL-1.3c", "deprecated": False},
    "lzma-sdk-9.11-to-9.20": {"id": "LZMA-SDK-9.11-to-9.20", "deprecated": False},
    "lzma-sdk-9.22": {"id": "LZMA-SDK-9.22", "deprecated": False},
    "makeindex": {"id": "MakeIndex", "deprecated": False},
    "minpack": {"id": "Minpack", "deprecated": False},
    "miros": {"id": "MirOS", "deprecated": False},
    "mit": {"id": "MIT", "deprecated": False},
    "mit-0": {"id": "MIT-0", "deprecated": False},
    "mit-advertising": {"id": "MIT-advertising", "deprecated": False},
    "mit-cmu": {"id": "MIT-CMU", "deprecated": False},
    "mit-enna": {"id": "MIT-enna", "deprecated": False},
    "mit-feh": {"id": "MIT-feh", "deprecated": False},
    "mit-modern-variant": {"id": "MIT-Modern-Variant", "deprecated": False},
    "mit-open-group": {"id": "MIT-open-group", "deprecated": False},
    "mitnfa": {"id": "MITNFA", "deprecated": False},
    "motosoto": {"id": "Motosoto", "deprecated": False},
    "mpi-permissive": {"id": "mpi-permissive", "deprecated": False},
    "mpich2": {"id": "mpich2", "deprecated": False},
    "mpl-1.0": {"id": "MPL-1.0", "deprecated": False},
    "mpl-1.1": {"id": "MPL-1.1", "deprecated": False},
    "mpl-2.0": {"id": "MPL-2.0", "deprecated": False},
    "mpl-2.0-no-copyleft-exception": {
        "id": "MPL-2.0-no-copyleft-exception",
        "deprecated": False,
    },
    "mplus": {"id": "mplus", "deprecated": False},
    "ms-lpl": {"id": "MS-LPL", "deprecated": False},
    "ms-pl": {"id": "MS-PL", "deprecated": False},
    "ms-rl": {"id": "MS-RL", "deprecated": False},
    "mtll": {"id": "MTLL", "deprecated": False},
    "mulanpsl-1.0": {"id": "MulanPSL-1.0", "deprecated": False},
    "mulanpsl-2.0": {"id": "MulanPSL-2.0", "deprecated": False},
    "multics": {"id": "Multics", "deprecated": False},
    "mup": {"id": "Mup", "deprecated": False},
    "naist-2003": {"id": "NAIST-2003", "deprecated": False},
    "nasa-1.3": {"id": "NASA-1.3", "deprecated": False},
    "naumen": {"id": "Naumen", "deprecated": False},
    "nbpl-1.0": {"id": "NBPL-1.0", "deprecated": False},
    "ncgl-uk-2.0": {"id": "NCGL-UK-2.0", "deprecated": False},
    "ncsa": {"id": "NCSA", "deprecated": False},
    "net-snmp": {"id": "Net-SNMP", "deprecated": False},
    "netcdf": {"id": "NetCDF", "deprecated": False},
    "newsletr": {"id": "Newsletr", "deprecated": False},
    "ngpl": {"id": "NGPL", "deprecated": False},
    "nicta-1.0": {"id": "NICTA-1.0", "deprecated": False},
    "nist-pd": {"id": "NIST-PD", "deprecated": False},
    "nist-pd-fallback": {"id": "NIST-PD-fallback", "deprecated": False},
    "nlod-1.0": {"id": "NLOD-1.0", "deprecated": False},
    "nlod-2.0": {"id": "NLOD-2.0", "deprecated": False},
    "nlpl": {"id": "NLPL", "deprecated": False},
    "nokia": {"id": "Nokia", "deprecated": False},
    "nosl": {"id": "NOSL", "deprecated": False},
    "noweb": {"id": "Noweb", "deprecated": False},
    "npl-1.0": {"id": "NPL-1.0", "deprecated": False},
    "npl-1.1": {"id": "NPL-1.1", "deprecated": False},
    "nposl-3.0": {"id": "NPOSL-3.0", "deprecated": False},
    "nrl": {"id": "NRL", "deprecated": False},
    "ntp": {"id": "NTP", "deprecated": False},
    "ntp-0": {"id": "NTP-0", "deprecated": False},
    "nunit": {"id": "Nunit", "deprecated": True},
    "o-uda-1.0": {"id": "O-UDA-1.0", "deprecated": False},
    "occt-pl": {"id": "OCCT-PL", "deprecated": False},
    "oclc-2.0": {"id": "OCLC-2.0", "deprecated": False},
    "odbl-1.0": {"id": "ODbL-1.0", "deprecated": False},
    "odc-by-1.0": {"id": "ODC-By-1.0", "deprecated": False},
    "ofl-1.0": {"id": "OFL-1.0", "deprecated": False},
    "ofl-1.0-no-rfn": {"id": "OFL-1.0-no-RFN", "deprecated": False},
    "ofl-1.0-rfn": {"id": "OFL-1.0-RFN", "deprecated": False},
    "ofl-1.1": {"id": "OFL-1.1", "deprecated": False},
    "ofl-1.1-no-rfn": {"id": "OFL-1.1-no-RFN", "deprecated": False},
    "ofl-1.1-rfn": {"id": "OFL-1.1-RFN", "deprecated": False},
    "ogc-1.0": {"id": "OGC-1.0", "deprecated": False},
    "ogdl-taiwan-1.0": {"id": "OGDL-Taiwan-1.0", "deprecated": False},
    "ogl-canada-2.0": {"id": "OGL-Canada-2.0", "deprecated": False},
    "ogl-uk-1.0": {"id": "OGL-UK-1.0", "deprecated": False},
    "ogl-uk-2.0": {"id": "OGL-UK-2.0", "deprecated": False},
    "ogl-uk-3.0": {"id": "OGL-UK-3.0", "deprecated": False},
    "ogtsl": {"id": "OGTSL", "deprecated": False},
    "oldap-1.1": {"id": "OLDAP-1.1", "deprecated": False},
    "oldap-1.2": {"id": "OLDAP-1.2", "deprecated": False},
    "oldap-1.3": {"id": "OLDAP-1.3", "deprecated": False},
    "oldap-1.4": {"id": "OLDAP-1.4", "deprecated": False},
    "oldap-2.0": {"id": "OLDAP-2.0", "deprecated": False},
    "oldap-2.0.1": {"id": "OLDAP-2.0.1", "deprecated": False},
    "oldap-2.1": {"id": "OLDAP-2.1", "deprecated": False},
    "oldap-2.2": {"id": "OLDAP-2.2", "deprecated": False},
    "oldap-2.2.1": {"id": "OLDAP-2.2.1", "deprecated": False},
    "oldap-2.2.2": {"id": "OLDAP-2.2.2", "deprecated": False},
    "oldap-2.3": {"id": "OLDAP-2.3", "deprecated": False},
    "oldap-2.4": {"id": "OLDAP-2.4", "deprecated": False},
    "oldap-2.5": {"id": "OLDAP-2.5", "deprecated": False},
    "oldap-2.6": {"id": "OLDAP-2.6", "deprecated": False},
    "oldap-2.7": {"id": "OLDAP-2.7", "deprecated": False},
    "oldap-2.8": {"id": "OLDAP-2.8", "deprecated": False},
    "oml": {"id": "OML", "deprecated": False},
    "openssl": {"id": "OpenSSL", "deprecated": False},
    "opl-1.0": {"id": "OPL-1.0", "deprecated": False},
    "opubl-1.0": {"id": "OPUBL-1.0", "deprecated": False},
    "oset-pl-2.1": {"id": "OSET-PL-2.1", "deprecated": False},
    "osl-1.0": {"id": "OSL-1.0", "deprecated": False},
    "osl-1.1": {"id": "OSL-1.1", "deprecated": False},
    "osl-2.0": {"id": "OSL-2.0", "deprecated": False},
    "osl-2.1": {"id": "OSL-2.1", "deprecated": False},
    "osl-3.0": {"id": "OSL-3.0", "deprecated": False},
    "parity-6.0.0": {"id": "Parity-6.0.0", "deprecated": False},
    "parity-7.0.0": {"id": "Parity-7.0.0", "deprecated": False},
    "pddl-1.0": {"id": "PDDL-1.0", "deprecated": False},
    "php-3.0": {"id": "PHP-3.0", "deprecated": False},
    "php-3.01": {"id": "PHP-3.01", "deprecated": False},
    "plexus": {"id": "Plexus", "deprecated": False},
    "polyform-noncommercial-1.0.0": {
        "id": "PolyForm-Noncommercial-1.0.0",
        "deprecated": False,
    },
    "polyform-small-business-1.0.0": {
        "id": "PolyForm-Small-Business-1.0.0",
        "deprecated": False,
    },
    "postgresql": {"id": "PostgreSQL", "deprecated": False},
    "psf-2.0": {"id": "PSF-2.0", "deprecated": False},
    "psfrag": {"id": "psfrag", "deprecated": False},
    "psutils": {"id": "psutils", "deprecated": False},
    "python-2.0": {"id": "Python-2.0", "deprecated": False},
    "python-2.0.1": {"id": "Python-2.0.1", "deprecated": False},
    "qhull": {"id": "Qhull", "deprecated": False},
    "qpl-1.0": {"id": "QPL-1.0", "deprecated": False},
    "rdisc": {"id": "Rdisc", "deprecated": False},
    "rhecos-1.1": {"id": "RHeCos-1.1", "deprecated": False},
    "rpl-1.1": {"id": "RPL-1.1", "deprecated": False},
    "rpl-1.5": {"id": "RPL-1.5", "deprecated": False},
    "rpsl-1.0": {"id": "RPSL-1.0", "deprecated": False},
    "rsa-md": {"id": "RSA-MD", "deprecated": False},
    "rscpl": {"id": "RSCPL", "deprecated": False},
    "ruby": {"id": "Ruby", "deprecated": False},
    "sax-pd": {"id": "SAX-PD", "deprecated": False},
    "saxpath": {"id": "Saxpath", "deprecated": False},
    "scea": {"id": "SCEA", "deprecated": False},
    "schemereport": {"id": "SchemeReport", "deprecated": False},
    "sendmail": {"id": "Sendmail", "deprecated": False},
    "sendmail-8.23": {"id": "Sendmail-8.23", "deprecated": False},
    "sgi-b-1.0": {"id": "SGI-B-1.0", "deprecated": False},
    "sgi-b-1.1": {"id": "SGI-B-1.1", "deprecated": False},
    "sgi-b-2.0": {"id": "SGI-B-2.0", "deprecated": False},
    "shl-0.5": {"id": "SHL-0.5", "deprecated": False},
    "shl-0.51": {"id": "SHL-0.51", "deprecated": False},
    "simpl-2.0": {"id": "SimPL-2.0", "deprecated": False},
    "sissl": {"id": "SISSL", "deprecated": False},
    "sissl-1.2": {"id": "SISSL-1.2", "deprecated": False},
    "sleepycat": {"id": "Sleepycat", "deprecated": False},
    "smlnj": {"id": "SMLNJ", "deprecated": False},
    "smppl": {"id": "SMPPL", "deprecated": False},
    "snia": {"id": "SNIA", "deprecated": False},
    "spencer-86": {"id": "Spencer-86", "deprecated": False},
    "spencer-94": {"id": "Spencer-94", "deprecated": False},
    "spencer-99": {"id": "Spencer-99", "deprecated": False},
    "spl-1.0": {"id": "SPL-1.0", "deprecated": False},
    "ssh-openssh": {"id": "SSH-OpenSSH", "deprecated": False},
    "ssh-short": {"id": "SSH-short", "deprecated": False},
    "sspl-1.0": {"id": "SSPL-1.0", "deprecated": False},
    "standardml-nj": {"id": "StandardML-NJ", "deprecated": True},
    "sugarcrm-1.1.3": {"id": "SugarCRM-1.1.3", "deprecated": False},
    "swl": {"id": "SWL", "deprecated": False},
    "tapr-ohl-1.0": {"id": "TAPR-OHL-1.0", "deprecated": False},
    "tcl": {"id": "TCL", "deprecated": False},
    "tcp-wrappers": {"id": "TCP-wrappers", "deprecated": False},
    "tmate": {"id": "TMate", "deprecated": False},
    "torque-1.1": {"id": "TORQUE-1.1", "deprecated": False},
    "tosl": {"id": "TOSL", "deprecated": False},
    "tu-berlin-1.0": {"id": "TU-Berlin-1.0", "deprecated": False},
    "tu-berlin-2.0": {"id": "TU-Berlin-2.0", "deprecated": False},
    "ucl-1.0": {"id": "UCL-1.0", "deprecated": False},
    "unicode-dfs-2015": {"id": "Unicode-DFS-2015", "deprecated": False},
    "unicode-dfs-2016": {"id": "Unicode-DFS-2016", "deprecated": False},
    "unicode-tou": {"id": "Unicode-TOU", "deprecated": False},
    "unlicense": {"id": "Unlicense", "deprecated": False},
    "upl-1.0": {"id": "UPL-1.0", "deprecated": False},
    "vim": {"id": "Vim", "deprecated": False},
    "vostrom": {"id": "VOSTROM", "deprecated": False},
    "vsl-1.0": {"id": "VSL-1.0", "deprecated": False},
    "w3c": {"id": "W3C", "deprecated": False},
    "w3c-19980720": {"id": "W3C-19980720", "deprecated": False},
    "w3c-20150513": {"id": "W3C-20150513", "deprecated": False},
    "watcom-1.0": {"id": "Watcom-1.0", "deprecated": False},
    "wsuipa": {"id": "Wsuipa", "deprecated": False},
    "wtfpl": {"id": "WTFPL", "deprecated": False},
    "wxwindows": {"id": "wxWindows", "deprecated": True},
    "x11": {"id": "X11", "deprecated": False},
    "x11-distribute-modifications-variant": {
        "id": "X11-distribute-modifications-variant",
        "deprecated": False,
    },
    "xerox": {"id": "Xerox", "deprecated": False},
    "xfree86-1.1": {"id": "XFree86-1.1", "deprecated": False},
    "xinetd": {"id": "xinetd", "deprecated": False},
    "xnet": {"id": "Xnet", "deprecated": False},
    "xpp": {"id": "xpp", "deprecated": False},
    "xskat": {"id": "XSkat", "deprecated": False},
    "ypl-1.0": {"id": "YPL-1.0", "deprecated": False},
    "ypl-1.1": {"id": "YPL-1.1", "deprecated": False},
    "zed": {"id": "Zed", "deprecated": False},
    "zend-2.0": {"id": "Zend-2.0", "deprecated": False},
    "zimbra-1.3": {"id": "Zimbra-1.3", "deprecated": False},
    "zimbra-1.4": {"id": "Zimbra-1.4", "deprecated": False},
    "zlib": {"id": "Zlib", "deprecated": False},
    "zlib-acknowledgement": {"id": "zlib-acknowledgement", "deprecated": False},
    "zpl-1.1": {"id": "ZPL-1.1", "deprecated": False},
    "zpl-2.0": {"id": "ZPL-2.0", "deprecated": False},
    "zpl-2.1": {"id": "ZPL-2.1", "deprecated": False},
}

EXCEPTIONS: dict[str, dict[str, str | bool]] = {
    "389-exception": {"id": "389-exception", "deprecated": False},
    "autoconf-exception-2.0": {"id": "Autoconf-exception-2.0", "deprecated": False},
    "autoconf-exception-3.0": {"id": "Autoconf-exception-3.0", "deprecated": False},
    "bison-exception-2.2": {"id": "Bison-exception-2.2", "deprecated": False},
    "bootloader-exception": {"id": "Bootloader-exception", "deprecated": False},
    "classpath-exception-2.0": {"id": "Classpath-exception-2.0", "deprecated": False},
    "clisp-exception-2.0": {"id": "CLISP-exception-2.0", "deprecated": False},
    "digirule-foss-exception": {"id": "DigiRule-FOSS-exception", "deprecated": False},
    "ecos-exception-2.0": {"id": "eCos-exception-2.0", "deprecated": False},
    "fawkes-runtime-exception": {"id": "Fawkes-Runtime-exception", "deprecated": False},
    "fltk-exception": {"id": "FLTK-exception", "deprecated": False},
    "font-exception-2.0": {"id": "Font-exception-2.0", "deprecated": False},
    "freertos-exception-2.0": {"id": "freertos-exception-2.0", "deprecated": False},
    "gcc-exception-2.0": {"id": "GCC-exception-2.0", "deprecated": False},
    "gcc-exception-3.1": {"id": "GCC-exception-3.1", "deprecated": False},
    "gnu-javamail-exception": {"id": "gnu-javamail-exception", "deprecated": False},
    "gpl-3.0-linking-exception": {
        "id": "GPL-3.0-linking-exception",
        "deprecated": False,
    },
    "gpl-3.0-linking-source-exception": {
        "id": "GPL-3.0-linking-source-exception",
        "deprecated": False,
    },
    "gpl-cc-1.0": {"id": "GPL-CC-1.0", "deprecated": False},
    "gstreamer-exception-2005": {"id": "GStreamer-exception-2005", "deprecated": False},
    "gstreamer-exception-2008": {"id": "GStreamer-exception-2008", "deprecated": False},
    "i2p-gpl-java-exception": {"id": "i2p-gpl-java-exception", "deprecated": False},
    "kicad-libraries-exception": {
        "id": "KiCad-libraries-exception",
        "deprecated": False,
    },
    "lgpl-3.0-linking-exception": {
        "id": "LGPL-3.0-linking-exception",
        "deprecated": False,
    },
    "libtool-exception": {"id": "Libtool-exception", "deprecated": False},
    "linux-syscall-note": {"id": "Linux-syscall-note", "deprecated": False},
    "llvm-exception": {"id": "LLVM-exception", "deprecated": False},
    "lzma-exception": {"id": "LZMA-exception", "deprecated": False},
    "mif-exception": {"id": "mif-exception", "deprecated": False},
    "nokia-qt-exception-1.1": {"id": "Nokia-Qt-exception-1.1", "deprecated": True},
    "ocaml-lgpl-linking-exception": {
        "id": "OCaml-LGPL-linking-exception",
        "deprecated": False,
    },
    "occt-exception-1.0": {"id": "OCCT-exception-1.0", "deprecated": False},
    "openjdk-assembly-exception-1.0": {
        "id": "OpenJDK-assembly-exception-1.0",
        "deprecated": False,
    },
    "openvpn-openssl-exception": {
        "id": "openvpn-openssl-exception",
        "deprecated": False,
    },
    "ps-or-pdf-font-exception-20170817": {
        "id": "PS-or-PDF-font-exception-20170817",
        "deprecated": False,
    },
    "qt-gpl-exception-1.0": {"id": "Qt-GPL-exception-1.0", "deprecated": False},
    "qt-lgpl-exception-1.1": {"id": "Qt-LGPL-exception-1.1", "deprecated": False},
    "qwt-exception-1.0": {"id": "Qwt-exception-1.0", "deprecated": False},
    "shl-2.0": {"id": "SHL-2.0", "deprecated": False},
    "shl-2.1": {"id": "SHL-2.1", "deprecated": False},
    "swift-exception": {"id": "Swift-exception", "deprecated": False},
    "u-boot-exception-2.0": {"id": "u-boot-exception-2.0", "deprecated": False},
    "universal-foss-exception-1.0": {
        "id": "Universal-FOSS-exception-1.0",
        "deprecated": False,
    },
    "wxwindows-exception-3.1": {"id": "WxWindows-exception-3.1", "deprecated": False},
    "x11vnc-openssl-exception": {"id": "x11vnc-openssl-exception", "deprecated": False},
}



DEFAULT_BUILD_DIRECTORY = "dist"

EXCLUDED_DIRECTORIES = frozenset(
    (
        # Python bytecode
        "__pycache__",
        # Git
        ".git",
        # Mercurial
        ".hg",
        # Puffling
        ".puffling",
        # tox
        ".tox",
        # nox
        ".nox",
        # Ruff
        ".ruff_cache",
        # pytest
        ".pytest_cache",
        # Mypy
        ".mypy_cache",
    )
)


class BuildEnvVars:
    LOCATION = "HATCH_BUILD_LOCATION"
    HOOKS_ONLY = "HATCH_BUILD_HOOKS_ONLY"
    NO_HOOKS = "HATCH_BUILD_NO_HOOKS"
    HOOKS_ENABLE = "HATCH_BUILD_HOOKS_ENABLE"
    HOOK_ENABLE_PREFIX = "HATCH_BUILD_HOOK_ENABLE_"
    CLEAN = "HATCH_BUILD_CLEAN"
    CLEAN_HOOKS_AFTER = "HATCH_BUILD_CLEAN_HOOKS_AFTER"


EDITABLES_REQUIREMENT = "editables~=0.3"



from __future__ import annotations

from typing import TYPE_CHECKING

from puffling.plugin import hookimpl
from puffling.version.source.code import CodeSource
from puffling.version.source.env import EnvSource
from puffling.version.source.regex import RegexSource

if TYPE_CHECKING:
    from puffling.version.source.plugin.interface import VersionSourceInterface


@hookimpl
def puffling_register_version_source() -> list[type[VersionSourceInterface]]:
    return [CodeSource, EnvSource, RegexSource]



from __future__ import annotations

import gzip
import json
import os
import tarfile
import tempfile
from contextlib import closing
from copy import copy
from io import BytesIO
from time import time as get_current_timestamp
from typing import TYPE_CHECKING, Any, Callable

from puffling.builders.config import BuilderConfig
from puffling.builders.plugin.interface import BuilderInterface
from puffling.builders.utils import (
    get_reproducible_timestamp,
    normalize_archive_path,
    normalize_file_permissions,
    normalize_relative_path,
    replace_file,
)
from puffling.metadata.spec import (
    DEFAULT_METADATA_VERSION,
    get_core_metadata_constructors,
)
from puffling.utils.constants import DEFAULT_BUILD_SCRIPT, DEFAULT_CONFIG_FILE

if TYPE_CHECKING:
    from types import TracebackType


class SdistArchive:
    def __init__(self, name: str, *, reproducible: bool) -> None:
        """
        https://peps.python.org/pep-0517/#source-distributions
        """
        self.name = name
        self.reproducible = reproducible
        self.timestamp: int | None = (
            get_reproducible_timestamp() if reproducible else None
        )

        raw_fd, self.path = tempfile.mkstemp(suffix=".tar.gz")
        self.fd = os.fdopen(raw_fd, "w+b")
        self.gz = gzip.GzipFile(fileobj=self.fd, mode="wb", mtime=self.timestamp)
        self.tf = tarfile.TarFile(fileobj=self.gz, mode="w", format=tarfile.PAX_FORMAT)
        self.gettarinfo = lambda *args, **kwargs: self.normalize_tar_metadata(
            self.tf.gettarinfo(*args, **kwargs)
        )

    def create_file(self, contents: str | bytes, *relative_paths: str) -> None:
        if not isinstance(contents, bytes):
            contents = contents.encode("utf-8")
        tar_info = tarfile.TarInfo(
            normalize_archive_path(os.path.join(self.name, *relative_paths))
        )
        tar_info.size = len(contents)
        if self.reproducible and self.timestamp is not None:
            tar_info.mtime = self.timestamp
        else:
            tar_info.mtime = int(get_current_timestamp())

        with closing(BytesIO(contents)) as buffer:
            self.tf.addfile(tar_info, buffer)

    def normalize_tar_metadata(self, tar_info: tarfile.TarInfo) -> tarfile.TarInfo:
        if not self.reproducible:
            return tar_info

        tar_info = copy(tar_info)
        tar_info.uid = 0
        tar_info.gid = 0
        tar_info.uname = ""
        tar_info.gname = ""
        tar_info.mode = normalize_file_permissions(tar_info.mode)
        if self.timestamp is not None:
            tar_info.mtime = self.timestamp

        return tar_info

    def __getattr__(self, name: str) -> Any:
        attr = getattr(self.tf, name)
        setattr(self, name, attr)
        return attr

    def __enter__(self) -> SdistArchive:  # noqa: PYI034
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self.tf.close()
        self.gz.close()
        self.fd.close()


class SdistBuilderConfig(BuilderConfig):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.__core_metadata_constructor: Callable[..., str] | None = None
        self.__strict_naming: bool | None = None
        self.__support_legacy: bool | None = None

    @property
    def core_metadata_constructor(self) -> Callable[..., str]:
        if self.__core_metadata_constructor is None:
            core_metadata_version = self.target_config.get(
                "core-metadata-version", DEFAULT_METADATA_VERSION
            )
            if not isinstance(core_metadata_version, str):
                message = f"Field `tool.puffling.build.targets.{self.plugin_name}.core-metadata-version` must be a string"
                raise TypeError(message)

            constructors = get_core_metadata_constructors()
            if core_metadata_version not in constructors:
                message = (
                    f'Unknown metadata version `{core_metadata_version}` for field '
                    f'`tool.puffling.build.targets.{self.plugin_name}.core-metadata-version`. '
                    f'Available: {", ".join(sorted(constructors))}'
                )
                raise ValueError(message)

            self.__core_metadata_constructor = constructors[core_metadata_version]

        return self.__core_metadata_constructor

    @property
    def strict_naming(self) -> bool:
        if self.__strict_naming is None:
            if "strict-naming" in self.target_config:
                strict_naming = self.target_config["strict-naming"]
                if not isinstance(strict_naming, bool):
                    message = f"Field `tool.puffling.build.targets.{self.plugin_name}.strict-naming` must be a boolean"
                    raise TypeError(message)
            else:
                strict_naming = self.build_config.get("strict-naming", True)
                if not isinstance(strict_naming, bool):
                    message = (
                        "Field `tool.puffling.build.strict-naming` must be a boolean"
                    )
                    raise TypeError(message)

            self.__strict_naming = strict_naming

        return self.__strict_naming

    @property
    def support_legacy(self) -> bool:
        if self.__support_legacy is None:
            self.__support_legacy = bool(
                self.target_config.get("support-legacy", False)
            )

        return self.__support_legacy


class SdistBuilder(BuilderInterface):
    """
    Build an archive of the source files
    """

    PLUGIN_NAME = "sdist"

    def get_version_api(self) -> dict[str, Callable]:
        return {"standard": self.build_standard}

    def get_default_versions(self) -> list[str]:  # noqa: PLR6301
        return ["standard"]

    def clean(  # noqa: PLR6301
        self,
        directory: str,
        versions: list[str],  # noqa: ARG002
    ) -> None:
        for filename in os.listdir(directory):
            if filename.endswith(".tar.gz"):
                os.remove(os.path.join(directory, filename))

    def build_standard(self, directory: str, **build_data: Any) -> str:
        found_packages = set()

        with SdistArchive(
            self.artifact_project_id, reproducible=self.config.reproducible
        ) as archive:
            for included_file in self.recurse_included_files():
                if self.config.support_legacy:
                    possible_package, file_name = os.path.split(
                        included_file.relative_path
                    )
                    if file_name == "__init__.py":
                        found_packages.add(possible_package)

                tar_info = archive.gettarinfo(
                    included_file.path,
                    arcname=normalize_archive_path(
                        os.path.join(
                            self.artifact_project_id, included_file.distribution_path
                        )
                    ),
                )

                if tar_info.isfile():
                    with open(included_file.path, "rb") as f:
                        archive.addfile(tar_info, f)
                else:  # no cov
                    # TODO: Investigate if this is necessary (for symlinks, etc.)
                    archive.addfile(tar_info)

            archive.create_file(
                self.config.core_metadata_constructor(
                    self.metadata, extra_dependencies=build_data["dependencies"]
                ),
                "PKG-INFO",
            )

            if self.config_settings:
                archive.create_file(
                    json.dumps(self.config_settings, indent=2),
                    "CONFIG-SETTINGS",
                )

            if self.config.support_legacy:
                archive.create_file(
                    self.construct_setup_py_file(
                        sorted(found_packages),
                        extra_dependencies=build_data["dependencies"],
                    ),
                    "setup.py",
                )

        target = os.path.join(directory, f"{self.artifact_project_id}.tar.gz")

        replace_file(archive.path, target)
        return target

    @property
    def artifact_project_id(self) -> str:
        return (
            self.project_id
            if self.config.strict_naming
            else f"{self.normalize_file_name_component(self.metadata.core.raw_name)}-{self.metadata.version}"
        )

    def construct_setup_py_file(
        self, packages: list[str], extra_dependencies: tuple[()] = ()
    ) -> str:
        contents = "# -*- coding: utf-8 -*-\nfrom setuptools import setup\n\n"

        contents += "setup(\n"

        contents += f"    name={self.metadata.core.name!r},\n"
        contents += f"    version={self.metadata.version!r},\n"

        if self.metadata.core.description:
            contents += f"    description={self.metadata.core.description!r},\n"

        if self.metadata.core.readme:
            contents += f"    long_description={self.metadata.core.readme!r},\n"

        authors_data = self.metadata.core.authors_data
        if authors_data["name"]:
            contents += f"    author={', '.join(authors_data['name'])!r},\n"
        if authors_data["email"]:
            contents += f"    author_email={', '.join(authors_data['email'])!r},\n"

        maintainers_data = self.metadata.core.maintainers_data
        if maintainers_data["name"]:
            contents += f"    maintainer={', '.join(maintainers_data['name'])!r},\n"
        if maintainers_data["email"]:
            contents += (
                f"    maintainer_email={', '.join(maintainers_data['email'])!r},\n"
            )

        if self.metadata.core.classifiers:
            contents += "    classifiers=[\n"

            for classifier in self.metadata.core.classifiers:
                contents += f"        {classifier!r},\n"

            contents += "    ],\n"

        dependencies = list(self.metadata.core.dependencies)
        dependencies.extend(extra_dependencies)
        if dependencies:
            contents += "    install_requires=[\n"

            for raw_specifier in dependencies:
                specifier = raw_specifier.replace("'", '"')
                contents += f"        {specifier!r},\n"

            contents += "    ],\n"

        if self.metadata.core.optional_dependencies:
            contents += "    extras_require={\n"

            for option, specifiers in self.metadata.core.optional_dependencies.items():
                if not specifiers:
                    continue

                contents += f"        {option!r}: [\n"

                for raw_specifier in specifiers:
                    specifier = raw_specifier.replace("'", '"')
                    contents += f"            {specifier!r},\n"

                contents += "        ],\n"

            contents += "    },\n"

        if (
            self.metadata.core.scripts
            or self.metadata.core.gui_scripts
            or self.metadata.core.entry_points
        ):
            contents += "    entry_points={\n"

            if self.metadata.core.scripts:
                contents += "        'console_scripts': [\n"

                for name, object_ref in self.metadata.core.scripts.items():
                    contents += f"            '{name} = {object_ref}',\n"

                contents += "        ],\n"

            if self.metadata.core.gui_scripts:
                contents += "        'gui_scripts': [\n"

                for name, object_ref in self.metadata.core.gui_scripts.items():
                    contents += f"            '{name} = {object_ref}',\n"

                contents += "        ],\n"

            if self.metadata.core.entry_points:
                for group, entry_points in self.metadata.core.entry_points.items():
                    contents += f"        {group!r}: [\n"

                    for name, object_ref in entry_points.items():
                        contents += f"            '{name} = {object_ref}',\n"

                    contents += "        ],\n"

            contents += "    },\n"

        if packages:
            src_layout = False
            contents += "    packages=[\n"

            for package in packages:
                if package.startswith(f"src{os.sep}"):
                    src_layout = True
                    contents += f"        {package.replace(os.sep, '.')[4:]!r},\n"
                else:
                    contents += f"        {package.replace(os.sep, '.')!r},\n"

            contents += "    ],\n"

            if src_layout:
                contents += "    package_dir={'': 'src'},\n"

        contents += ")\n"

        return contents

    def get_default_build_data(self) -> dict[str, Any]:
        force_include = {}
        for filename in ["pyproject.toml", DEFAULT_CONFIG_FILE, DEFAULT_BUILD_SCRIPT]:
            path = os.path.join(self.root, filename)
            if os.path.exists(path):
                force_include[path] = filename
        build_data = {"force_include": force_include, "dependencies": []}

        for exclusion_files in self.config.vcs_exclusion_files.values():
            for exclusion_file in exclusion_files:
                force_include[exclusion_file] = os.path.basename(exclusion_file)

        readme_path = self.metadata.core.readme_path
        if readme_path:
            readme_path = normalize_relative_path(readme_path)
            force_include[os.path.join(self.root, readme_path)] = readme_path

        license_files = self.metadata.core.license_files
        if license_files:
            for license_file in license_files:
                relative_path = normalize_relative_path(license_file)
                force_include[os.path.join(self.root, relative_path)] = relative_path

        return build_data

    @classmethod
    def get_config_class(cls) -> type[SdistBuilderConfig]:
        return SdistBuilderConfig
